import { useState } from 'react'
import {
  Table,
  TableBody,
  TableCell,
  TableContainer,
  TableHead,
  TableRow,
  Paper,
  Accordion,
  AccordionSummary,
  AccordionDetails,
} from '@mui/material'
import ExpandMoreIcon from '@mui/icons-material/ExpandMore'

import Page from 'components/page'
import { useTranslation } from 'react-i18next'
import { Link } from 'react-router-dom'
import { Guest } from 'interfaces'
import SponsorGuestButtons from '../../components/sponsorGuestButtons'

interface GuestProps {
  persons: Guest[]
}
interface PersonLineProps {
  person: Guest
}

const PersonLine = ({ person }: PersonLineProps) => {
  const [t, i18n] = useTranslation(['common'])

  return (
    <TableRow
      key={person.name}
      sx={{ '&:last-child td, &:last-child th': { border: 0 } }}
    >
      <TableCell component="th" scope="row">
        {person.name}
      </TableCell>
      <TableCell align="left">
        {i18n.language === 'en' ? person.role_en : person.role_nb}
      </TableCell>
      <TableCell align="left">{person.period}</TableCell>
      <TableCell align="left">
        {i18n.language === 'en' ? person.ou_en : person.ou_nb}
      </TableCell>
      <TableCell align="left">
        <Link to={`/sponsor/guest/${person.pid}`}>{t('common:details')}</Link>
      </TableCell>
    </TableRow>
  )
}

const ActiveGuests = ({ persons }: GuestProps) => {
  const [activeExpanded, setActiveExpanded] = useState(false)

  // Only show active people
  let guests = persons.length > 0 ? persons : []
  if (guests.length > 0) {
    guests = guests.filter((person) => person.active)
  }
  const [t] = useTranslation(['common'])
  return (
    <Accordion
      expanded={activeExpanded}
      onChange={() => {
        setActiveExpanded(!activeExpanded)
      }}
    >
      <AccordionSummary expandIcon={<ExpandMoreIcon />}>
        <h4>{t('common:activeGuests')}</h4>
      </AccordionSummary>
      <AccordionDetails>
        <p>{t('common:activeGuestsDescription')}</p>
        <TableContainer component={Paper}>
          <Table sx={{ minWidth: 650 }} aria-label="simple table">
            <TableHead sx={{ backgroundColor: 'primary.light' }}>
              <TableRow>
                <TableCell>{t('common:name')}</TableCell>
                <TableCell align="left">{t('common:role')}</TableCell>
                <TableCell align="left">{t('common:period')}</TableCell>
                <TableCell align="left">{t('common:ou')}</TableCell>
                <TableCell align="left">{t('common:choice')}</TableCell>
              </TableRow>
            </TableHead>
            <TableBody>
              {guests.map((person) => (
                <PersonLine person={person} />
              ))}

              <TableRow>
                <TableCell>
                  {guests.length > 0 ? '' : t('common:noActiveGuests')}
                </TableCell>
              </TableRow>
            </TableBody>
          </Table>
        </TableContainer>
      </AccordionDetails>
    </Accordion>
  )
}

const WaitingGuests = ({ persons }: GuestProps) => {
  const [waitingExpanded, setWaitingExpanded] = useState(false)

  // Only show non-active people
  let guests = persons.length > 0 ? persons : []
  if (guests.length > 0) {
    guests = guests.filter((person) => !person.active)
  }
  const [t] = useTranslation(['common'])

  return (
    <Accordion
      expanded={waitingExpanded}
      onChange={() => {
        setWaitingExpanded(!waitingExpanded)
      }}
      sx={{ border: 'none' }}
    >
      <AccordionSummary expandIcon={<ExpandMoreIcon />}>
        <h4>{t('common:waitingGuests')}</h4>
      </AccordionSummary>
      <AccordionDetails>
        <p>{t('common:waitingGuestsDescription')}</p>

        <TableContainer component={Paper}>
          <Table sx={{ minWidth: 650 }} aria-label="simple table">
            <TableHead sx={{ backgroundColor: 'primary.light' }}>
              <TableRow>
                <TableCell>{t('common:name')}</TableCell>
                <TableCell align="left">{t('common:role')}</TableCell>
                <TableCell align="left">{t('common:period')}</TableCell>
                <TableCell align="left">{t('common:ou')}</TableCell>
                <TableCell align="left">{t('common:choice')}</TableCell>
              </TableRow>
            </TableHead>
            <TableBody>
              {guests.map((person) => (
                <PersonLine person={person} />
              ))}
              <TableRow>
                <TableCell>
                  {guests.length > 0 ? '' : t('common:noWaitingGuests')}
                </TableCell>
              </TableRow>
            </TableBody>
          </Table>
        </TableContainer>
      </AccordionDetails>
    </Accordion>
  )
}

interface FrontPageProps {
  guests: Guest[]
}

function FrontPage({ guests }: FrontPageProps) {
  return (
    <Page>
      <SponsorGuestButtons yourGuestsActive />
      <WaitingGuests persons={guests} />
      <ActiveGuests persons={guests} />
    </Page>
  )
}

export default FrontPage
