import phonenumbers
from rest_framework import serializers

from greg.models import Identity, Person
from greg.utils import is_valid_norwegian_national_id_number


def _validateNorwegianNationalIdNumber(value):
    # Not excepted that D-numbers will be entered through the form, so only
    # accept national ID numbers
    if not is_valid_norwegian_national_id_number(value, False):
        raise serializers.ValidationError("Not a valid Norwegian national ID number")


def _validatePhoneNumber(value):
    if not phonenumbers.is_valid_number(phonenumbers.parse(value)):
        raise serializers.ValidationError("Invalid phone number")


class GuestRegisterSerializer(serializers.ModelSerializer):
    # TODO first_name and last_name set as not required to throwing an exception if they are not included in what is sent back from the frontend. It is perhaps not required that they are in the reponse from the client if the guest should be allowed to change them
    first_name = serializers.CharField(required=False)
    last_name = serializers.CharField(required=False)
    email = serializers.CharField(required=True)
    mobile_phone = serializers.CharField(
        required=True, validators=[_validatePhoneNumber]
    )
    fnr = serializers.CharField(
        required=False, validators=[_validateNorwegianNationalIdNumber]
    )

    def update(self, instance, validated_data):
        email = validated_data.pop("email")
        mobile_phone = validated_data.pop("mobile_phone")

        if not instance.private_email:
            Identity.objects.create(
                person=instance,
                type=Identity.IdentityType.PRIVATE_EMAIL,
                value=email,
            )
        else:
            private_email = instance.private_email
            private_email.value = email
            private_email.save()

        if not instance.private_mobile:
            Identity.objects.create(
                person=instance,
                type=Identity.IdentityType.PRIVATE_MOBILE_NUMBER,
                value=mobile_phone,
            )
        else:
            private_mobile = instance.private_mobile
            private_mobile.value = mobile_phone
            private_mobile.save()

        if "fnr" in validated_data:
            fnr = validated_data.pop("fnr")
            if not instance.fnr:
                Identity.objects.create(
                    person=instance,
                    type=Identity.IdentityType.NORWEGIAN_NATIONAL_ID_NUMBER,
                    value=fnr,
                )
            else:
                fnr_existing = instance.fnr
                fnr_existing.value = fnr
                fnr_existing.save()

        # TODO: we only want to allow changing the name if we don't have one
        #       from a reliable source (Feide/KORR)
        # TODO Comment back in after it is decided if name updates are allowed
        # instance.first_name = validated_data["first_name"]
        # instance.last_name = validated_data["last_name"]

        return instance

    class Meta:
        model = Person
        fields = ("id", "first_name", "last_name", "email", "mobile_phone", "fnr")
        read_only_fields = ("id",)
