import logging
import re
import time

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.backends import BaseBackend
from django.core.exceptions import SuspiciousOperation
from mozilla_django_oidc.auth import OIDCAuthenticationBackend

from django.utils import timezone

from greg.models import Identity, Person, Sponsor
from gregui.models import GregUserProfile

logger = logging.getLogger(__name__)


class DevBackend(BaseBackend):
    """
    Development backend, no password checking.

    TODO:
    - Expand to add person/sponsor for the user
    - How do we do login? A dev login site?

    """

    def __init__(self):
        self.UserModel = get_user_model()

    dev_users = {
        "bruker1": {
            "email": "bruker1@uio.no",
            "first_name": "test",
            "last_name": "testesen",
        },
        "bruker2": {
            "email": "bruker2@uio.no",
            "first_name": "foo",
            "last_name": "foosen",
        },
    }

    def get_userinfo(self, username):
        return self.dev_users.get(username, {})

    def get_or_create_user(self, username):
        if not username:
            return None
        try:
            user = self.UserModel.objects.get(username=username)
        except self.UserModel.DoesNotExist:
            userinfo = self.get_userinfo(username)
            user = self.UserModel(username=username, **userinfo)
            user.save()
        return user

    def authenticate(self, request, **kwargs):
        # This has been made so that developers can authenticate as any user in test
        username = kwargs.get("username")
        logger.debug("Username is %s", username)
        request.session["oidc_id_token_payload"] = {"iat": time.time()}
        return self.get_or_create_user(username)

    def get_user(self, user_id):
        try:
            return self.UserModel.objects.get(pk=user_id)
        except self.UserModel.DoesNotExist:
            return None


# This regex is quite naive since it assumes that the person only has one last_name,
# and everything before that is the first_name.
NAMES_REGEX = re.compile(r"^(?P<first_name>.+) (?P<last_name>.+)$")
ID_REGEX = re.compile(r"^(?P<id_type>[^:]+):(?P<id_value>.+)$")


def extract_userinfo(claims):
    name = claims["name"]
    match = re.search(NAMES_REGEX, name)
    first_name = match.group("first_name")
    last_name = match.group("last_name")
    userinfo = {
        "email": claims["email"],
        "first_name": first_name,
        "last_name": last_name,
    }
    secondary_user_ids = claims["connect-userid_sec"]
    for user_id in secondary_user_ids:
        match = re.search(ID_REGEX, user_id)
        id_type = match.group("id_type")
        id_value = match.group("id_value")
        if id_type == "feide":
            userinfo["userid_feide"] = id_value
        # TODO:
        #  Should we also save nin of the user?
        #  What should we do if a user has multiple feide_ids?
    return userinfo


class ValidatingOIDCBackend(OIDCAuthenticationBackend):
    def validate_issuer(self, payload):
        issuer = self.get_settings("OIDC_OP_ISSUER")
        if not issuer == payload["iss"]:
            raise SuspiciousOperation(
                '"iss": %r does not match configured value for OIDC_OP_ISSUER: %r'
                % (payload["iss"], issuer)
            )

    def validate_audience(self, payload):
        client_id = self.get_settings("OIDC_RP_CLIENT_ID")
        trusted_audiences = self.get_settings("OIDC_TRUSTED_AUDIENCES", [])
        trusted_audiences = set(trusted_audiences)
        trusted_audiences.add(client_id)

        audience = payload["aud"]
        if not isinstance(audience, list):
            audience = [audience]
        audience = set(audience)
        if client_id not in audience:
            raise SuspiciousOperation(
                "Client id not present in audiences: %r" % audience
            )
        distrusted_audiences = audience.difference(trusted_audiences)
        if distrusted_audiences:
            raise SuspiciousOperation(
                '"aud" contains distrusted audiences: %r' % distrusted_audiences
            )

    def validate_expiry(self, payload):
        expire_time = payload["exp"]
        now = time.time()
        if now > expire_time:
            raise SuspiciousOperation(
                "Id-token is expired %r > %r" % (now, expire_time)
            )

    def validate_id_token(self, payload):
        """Validate the content of the id token as required by OpenID Connect 1.0

        This aims to fulfill point 2. 3. and 9. under section 3.1.3.7. ID Token
        Validation
        """
        self.validate_issuer(payload)
        self.validate_audience(payload)
        self.validate_expiry(payload)
        return payload

    def verify_token(self, token, **kwargs):
        # Overriding this method to fix issues described here:
        # https://github.com/mozilla/mozilla-django-oidc/issues/340
        payload = super().verify_token(token, **kwargs)
        return self.validate_id_token(payload)


class GregOIDCBackend(ValidatingOIDCBackend):
    """
    Custom mozilla OIDC backend for Greg.

    TODO:
    - Support for id-porten fnr login.
    """

    def verify_token(self, token, **kwargs):
        payload = super().verify_token(token, **kwargs)
        # While mozilla_django_oidc does have support for automatically storing the
        # id-token, it is more useful for us to be able to store the actual payload of
        # the id-token.
        session = self.request.session
        session["oidc_id_token_payload"] = {"iat": payload["iat"]}
        return payload

    def get_username(self, claims):
        return "{}{}".format(self.get_settings("OIDC_OP_ISSUER"), claims)

    def filter_users_by_claims(self, claims):
        #  Ideally dataporten should include 'iss' in its claims, so that it can be
        #  used along with sub to create a unique identifier for the user. However, if
        #  the id-token has been validated, it means that the 'iss' specified by the
        #  id-token is equal to what is specified in the configuration (OIDC_OP_ISSUER).
        sub = claims.get("sub")
        if not sub:
            return self.UserModel.objects.none()
        username = self.get_username(sub)

        try:
            user = self.UserModel.objects.filter(username=username)
            return user
        except self.UserModel.DoesNotExist:
            return self.UserModel.objects.none()

    def get_or_create_user(self, access_token, id_token, payload):
        # This method has been overridden to update the user_profile object
        # for a user at login
        user = super().get_or_create_user(access_token, id_token, payload)
        if user:
            # Update or create a user_profile
            userinfo = extract_userinfo(
                self.get_userinfo(access_token, id_token, payload)
            )
            self._get_or_create_greg_user_profile(userinfo, user)
        return user

    def _get_or_create_person(self, userinfo):
        # Update any new person info
        person, _ = Person.objects.update_or_create(
            first_name=userinfo["first_name"],
            last_name=userinfo["last_name"],
        )
        person.save()
        return person

    def _get_or_create_greg_user_profile(self, userinfo, user):
        """
        Get or create a GregUserProfile.

        We try to match a logged inn user with an existing profile.
        If no profile exists, we check if the feide_id is registered.

        """
        try:
            user_profile = GregUserProfile.objects.get(user=user)
        except GregUserProfile.DoesNotExist:

            # Check if user is a sponsor
            try:
                sponsor = Sponsor.objects.get(feide_id=userinfo["userid_feide"])
            except Sponsor.DoesNotExist:
                sponsor = None

            try:
                # TODO, match against fnr if using id-porten.
                identity = Identity.objects.get(
                    type="feide_id", value=userinfo["userid_feide"]
                )
                person = identity.person
            except Identity.DoesNotExist:
                # Find or create person, and add identity
                person = self._get_or_create_person(userinfo)
                identity = Identity(
                    type=Identity.IdentityType.FEIDE_ID, value=userinfo["userid_feide"], person=person,
                    source=settings.FEIDE_SOURCE, verified=Identity.Verified.AUTOMATIC,
                    verified_at=timezone.now()
                )
                identity.save()

            try:
                Identity.objects.get(
                    type="private_email", value=userinfo["email"]
                )
            except Identity.DoesNotExist:
                # Add email if missing
                email_identity = Identity(
                    type="private_email", value=userinfo["email"], person=person, source=settings.FEIDE_SOURCE,
                )
                email_identity.save()

            user_profile = GregUserProfile(
                user=user,
                person=person,
                sponsor=sponsor,
                userid_feide=userinfo["userid_feide"],
            )
            user_profile.save()

        return user_profile

    def create_user(self, claims):
        userinfo = extract_userinfo(claims)
        username = self.get_username(claims["sub"])
        user = self.UserModel(
            username=username,
            first_name=userinfo["first_name"],
            last_name=userinfo["last_name"],
            email=userinfo["email"],
        )
        user.save()

        # Create a user_profile if missing
        self._get_or_create_greg_user_profile(userinfo, user)
        return user

    def update_user(self, user, claims):
        username = self.get_username(claims["sub"])
        user = self.UserModel.objects.get(username=username)
        userinfo = extract_userinfo(claims)
        for key, new_value in userinfo.items():
            current_value = getattr(user, key, None)
            if not new_value == current_value:
                setattr(user, key, new_value)
        user.save()
        # Create a user_profile if missing
        self._get_or_create_greg_user_profile(userinfo, user)
        return user


def provider_logout(request):
    redirect_url = settings.OIDC_END_SESSION_ENDPOINT
    return redirect_url
